#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "util.h"

// implementing the KM algorithm
// w: the weight matrix
// n: # of vertices on each side
// perm: max-weight matching, i --> perm[i], must be of size n
// return: total cost of the max-weight matching
double KM(double* w, int n, int* perm) {

	double *A, *B; // labels
	int *visited, *last;
	int *queue;
	int fr, re;
	int* _perm; // matching back

	assert(A = (double*)malloc(n * sizeof(double)));
	assert(B = (double*)malloc(n * sizeof(double)));
	assert(visited = (int*)malloc(2*n * sizeof(int)));
	assert(last = (int*)malloc(2*n * sizeof(int)));
	assert(queue = (int*)malloc(2*n * sizeof(int)));
	assert(_perm = (int*)malloc(n * sizeof(int)));

	// Initialization: A[i] = max w(i,j); B[i] = 0
	for(int i = 0; i < n; i++) {
		double maxw = -1e100;
		for(int j = 0; j < n; j++)
			if (w[ADDR2D(i, j, n)] > maxw) maxw = w[ADDR2D(i, j, n)];
		A[i] = maxw;
		B[i] = 0;
	}

	while (1) {

		// use BFS to find a perfect matching
		int unsuccess = -1;
		for(int i = 0; i < n; i++) {
			perm[i] = -1;
			_perm[i] = -1;
		}
		for(int root = 0; root < n; root++) {
			
			// Do BFS
			for(int i = 0; i < 2*n; i++) {
				visited[i] = 0;
				last[i] = -1;
			}
			fr = 0;
			re = 1;
			queue[fr] = root;
			visited[root] = 1;
			
			while (fr < re) {
				int v = queue[fr++];
				if (v < n) {
					// vertex on the left side
					int i = v;
					for(int j = 0; j < n; j++)
						if (!visited[n+j] && fabs(A[i] + B[j] - w[ADDR2D(i, j, n)]) < MINEPS) {
							visited[n+j] = 1;
							last[n+j] = i;
							queue[re++] = n+j;
							if (_perm[j] == -1) goto label0;
						}
				}
				else {
					// vertex on the right side
					int j = v - n;
					int i = _perm[j];
					if (!visited[i]) {
						visited[i] = 1;
						last[i] = n+j;
						queue[re++] = i;
					}
				}
			}
	
label0:

			if (fr < re) {
				// an augmentation path has been found: augment it, and continue the loop
				int v = queue[re-1];
				while (v != -1) {
					int lastv = last[v];
					perm[lastv] = v - n;
					_perm[v-n] = lastv;
					v = last[lastv];
				}
			}
			else {
				unsuccess = root;
				break;
			}

		}

		if (unsuccess == -1) {
			//successfully found a perfect matching
			break;
		}
		else {
			// alter the vertex labels
			double d = 1e100;
			for(int i = 0; i < n; i++)
				if (visited[i]) {
					for(int j = 0; j < n; j++)
						if (!visited[n+j]) {
							double value = A[i] + B[j] - w[ADDR2D(i, j, n)];
							if (value < d) d = value;
						}
				}
			for(int i = 0; i < n; i++) {
				if (visited[i]) A[i] -= d;
				if (visited[n+i]) B[i] += d;
			}
		}

	}

	free(A);
	free(B);
	free(visited);
	free(last);
	free(queue);
	free(_perm);

	double ret = 0;
	for(int i = 0; i < n; i++)
		ret += w[ADDR2D(i, perm[i], n)];
	return ret;

}

double dfs(double w[][10], int n, bool visited[], int v) {

	if (v == n)
		return 0;
	else {
		double ret = -1e100;
		for(int i = 0; i < n; i++) 
			if (!visited[i]) {
				visited[i] = true;
				double t = dfs(w, n, visited, v+1) + w[v][i];
				if (t > ret) ret = t;
				visited[i] = false;
			}
		return ret;
	}

}

int main() {

	double w[10][10];
	int perm[10];
	int n = 10;

	for(int i = 0; i < n; i++)
		for(int j = 0; j < n; j++) {
			w[i][j] = ((double)rand() / RAND_MAX) * 2 + 1;
		}


	double v = KM((double*)w, n, perm);
	printf("v = %lf\n", v);


	bool visited[10];
	memset(visited, 0, sizeof(visited));
	double ref_v = dfs(w, n, visited, 0);
	printf("ref_v = %lf\n", ref_v);
	return 0;

}
