#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <time.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "util.h"
#include "doc.h"
#include "slda.h"
#include "specslda.h"

void vec_normalize(gsl_vector* x, int n) {
	double norm = gsl_blas_dnrm2(x);
	for(int i = 0; i < n; i++)
		x->data[i] /= norm; 
}

gsl_matrix* whiten(gsl_matrix* M2, int V, int K) {

/*	gsl_matrix* UL = gsl_matrix_alloc(V, V);
	gsl_matrix* UR = gsl_matrix_alloc(V, V);
	gsl_vector* S = gsl_vector_alloc(V);

	gsl_vector* work = gsl_vector_alloc(V);
	gsl_matrix_memcpy(UL, M2);
	gsl_linalg_SV_decomp(UL, UR, S, work);
	gsl_vector_free(work);
	gsl_matrix_memcpy(UL, M2);
	gsl_linalg_SV_decomp_jacobi(UL, UR, S);

	// use redsvd to perform SVD
	FILE *f = fopen("M2.temp", "w");
	for(int i = 0; i < V; i++) {
		for(int j = 0; j < V; j++)
			fprintf(f, "%e ", M2->data[ADDR2D(i, j, V)]);
		fprintf(f, "\n");
	}
	fclose(f);
	char cmd[1000];
	memset(cmd, 0, sizeof(cmd));
	sprintf(cmd, "redsvd -i M2.temp -o M2.temp -r %d -f dense", V);
	system(cmd);
	f = fopen("M2.temp.U", "r");
	for(int i = 0; i < V; i++)
		for(int j = 0; j < V; j++)
			fscanf(f, "%lf", &(UL->data[ADDR2D(i, j, V)]));
	fclose(f);
	f = fopen("M2.temp.S", "r");
	for(int i = 0; i < V; i++)
		fscanf(f, "%lf", &(S->data[i]));
	fclose(f);

	gsl_matrix* W = gsl_matrix_alloc(V, K);
	for(int i = 0; i < V; i++)
		for(int j = 0; j < K; j++)
			W->data[ADDR2D(i, j, K)] = UL->data[ADDR2D(i, j, V)] / sqrt(S->data[j]);

	gsl_matrix_free(UL);
	gsl_matrix_free(UR);
	gsl_vector_free(S);*/

	gsl_matrix* W = gsl_matrix_alloc(V, K);
	FILE *f = fopen("whiten.input", "w");
	fprintf(f, "%d %d\n", V, K);
	for(int i = 0; i < V; i++) {
		for(int j = 0; j < V; j++)
			fprintf(f, "%0.12e ", M2->data[ADDR2D(i, j, V)]);
		fprintf(f, "\n");
	}
	fclose(f);
	system("matlab -nodisplay -nosplash -nodesktop -r whiten");
	f = fopen("whiten.output", "r");
	for(int i = 0; i < V; i++)
		for(int j = 0; j < K; j++)
			fscanf(f, "%lf", &(W->data[ADDR2D(i, j, K)]));
	fclose(f);

	// check
	{
		gsl_matrix* work1 = gsl_matrix_alloc(V, K);
		gsl_matrix_set_all(work1, 0);
		gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, M2, W, 0, work1);
		gsl_matrix* work2 = gsl_matrix_alloc(K, K);
		gsl_matrix_set_all(work2, 0);
		gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, W, work1, 0, work2);
		for(int i = 0; i < K; i++)
			work2->data[ADDR2D(i, i, K)] -= 1.0;
		double sum = 0;
		for(int i = 0; i < K; i++)
			for(int j = 0; j < K; j++)
				sum += SQR(work2->data[ADDR2D(i, j, K)]);
		printf("check: ||WM2W - I||_F = %e\n", sum);
		gsl_matrix_free(work1);
		gsl_matrix_free(work2);
	}

	return W;

}

double fnorm(double* A, int n) {
	double ret = 0;
	for(int i = 0; i < n; i++)
		for(int j = 0; j < n; j++)
			for(int k = 0; k < n; k++) {
				double value = A[ADDR3D(i, j, k, n)];
				ret += SQR(value);
			}
	return ret;
}

void robust_tensor_decomposition(double* A, int K, int V, int L, int T, gsl_vector* lambda, gsl_vector** v, gsl_rng* rng) {

	gsl_vector_set_all(lambda, 0);
	for(int k = 0; k < K; k++)
		gsl_vector_set_all(v[k], 0);

	gsl_vector* argmax_theta = gsl_vector_alloc(K);
	gsl_vector* theta = gsl_vector_alloc(K);
	gsl_vector* work = gsl_vector_alloc(K);
	double max_theta;

	for(int k = 0; k < K; k++) {

		printf("Recovering robust eigen-vector: k = %d, residual = %lf\n", k, fnorm(A, K));

		gsl_vector_set_all(argmax_theta, 0);
		max_theta = -1e100;

		for(int tau = 0; tau < L; tau++) {
			// draw theta uniformly random from the unit sphere in R^k
			for(int i = 0; i < K; i++)
				theta->data[i] = gsl_ran_gaussian(rng, 1.0);
			vec_normalize(theta, K);

			// Now, proceed ..
			for(int t = 0; t < T; t++) {
				//compute: theta = A(I, theta, theta)	
				gsl_vector_memcpy(work, theta);
				for(int i = 0; i < K; i++) {
					double sum = 0;
					for(int k2 = 0; k2 < K; k2++)
						for(int k3 = 0; k3 < K; k3++)
							sum += A[ADDR3D(i, k2, k3, K)] * work->data[k2] * work->data[k3];
					theta->data[i] = sum;
				}
				vec_normalize(theta, K);
			}

			// compute v_theta and compare it with argmax_theta
			double v_theta = 0;
			for(int k1 = 0; k1 < K; k1++)
				for(int k2 = 0; k2 < K; k2++)
					for(int k3 = 0; k3 < K; k3++) {
						v_theta += A[ADDR3D(k1, k2, k3, K)] * theta->data[k1] * theta->data[k2] * theta->data[k3];
					}
			if (v_theta > max_theta) {
				max_theta = v_theta;
				gsl_vector_memcpy(argmax_theta, theta);
			}
		}

		// Let's do another round of update on the argmax theta
		gsl_vector_memcpy(theta, argmax_theta);
		for(int t = 0; t < T; t++) {
			gsl_vector_memcpy(work, theta);
			for(int i = 0; i < K; i++) {
				double sum = 0;
				for(int k2 = 0; k2 < K; k2++)
					for(int k3 = 0; k3 < K; k3++)
						sum += A[ADDR3D(i, k2, k3, K)] * work->data[k2] * work->data[k3];
				theta->data[i] = sum;
			}
			vec_normalize(theta, K);
		}

		// Now, record v, lambda, and subtract this component from A
		gsl_vector_memcpy(v[k], theta);
		lambda->data[k] = 0;
		for(int k1 = 0; k1 < K; k1++)
			for(int k2 = 0; k2 < K; k2++)
				for(int k3 = 0; k3 < K; k3++) {
					lambda->data[k] += A[ADDR3D(k1, k2, k3, K)] * theta->data[k1] * theta->data[k2] * theta->data[k3];
				}

		for(int k1 = 0; k1 < K; k1++)
			for(int k2 = 0; k2 < K; k2++)
				for(int k3 = 0; k3 < K; k3++) {
					A[ADDR3D(k1, k2, k3, K)] -= lambda->data[k] * theta->data[k1] * theta->data[k2] * theta->data[k3];
				}

	}

	printf("final residual = %lf\n", fnorm(A, K));

	gsl_vector_free(work);
	gsl_vector_free(theta);
	gsl_vector_free(argmax_theta);

}

strct_slda* specslda_fit(strct_doc* docs, int ndocs, double alpha0, int K, int V, int L, int T) {

	gsl_rng_env_setup();
	gsl_rng* rng = gsl_rng_alloc(gsl_rng_default);

	int tm_record;
	double denom;
	int percentage;

	printf("ndocs = %d, K = %d, V = %d, L = %d, T = %d, alpha0 = %lf\n", ndocs, K, V, L, T, alpha0);

	puts("Computing 1-st order moments ...");
	tm_record = clock();
	gsl_vector* M1 = gsl_vector_alloc(V);
	gsl_vector_set_all(M1, 0);
	percentage = 0;
	for(int i = 0; i < ndocs; i++) {
		if (i > ndocs * 0.01 * (percentage + 1)) {
			putchar('.');
			fflush(stdout);
			percentage ++;
		}
		denom = 1.0 / docs[i].word_count;
		for(int j = 0; j < docs[i].item_count; j++) {
			int v = docs[i].index[j];
			int c = docs[i].count[j];
			M1->data[v] += c * denom;
		}	
	}
	puts("");
	denom = 1.0 / ndocs;
	for(int i = 0; i < V; i++)
		M1->data[i] *= denom;
	// check
	assert(fabs(gsl_blas_dasum(M1) - 1) < MINEPS);

	printf("time = %lf\n", (double)(clock() - tm_record) / CLOCKS_PER_SEC);

	puts("Computing 2-nd order moments ...");
	tm_record = clock();
	gsl_matrix* A2 = gsl_matrix_alloc(V, V);
	gsl_matrix_set_all(A2, 0);
	percentage = 0;
	for(int i = 0; i < ndocs; i++) {
		if (i > ndocs * 0.01 * (percentage + 1)) {
			putchar('.');
			fflush(stdout);
			percentage ++;
		}
		int m = docs[i].word_count;
		denom = 1.0 / (m * (m-1));
		for(int j1 = 0; j1 < docs[i].item_count; j1++)
			for(int j2 = 0; j2 < docs[i].item_count; j2++) {
				int v1 = docs[i].index[j1], v2 = docs[i].index[j2];
				int c1 = docs[i].count[j1], c2 = docs[i].count[j2];
				if (j1 == j2) {
					A2->data[ADDR2D(v1, v2, V)] += c1 * (c2-1) * denom;
				}
				else {
					A2->data[ADDR2D(v1, v2, V)] += c1 * c2 * denom;
				}
			}
	}
	puts("");
	denom = 1.0 / ndocs;
	for(int i = 0; i < V; i++)
		for(int j = 0; j < V; j++)
			A2->data[ADDR2D(i, j, V)] *= denom;
	// check
	{
		double sum = 0;
		for(int i = 0; i < V; i++)
			for(int j = 0; j < V; j++)
				sum += A2->data[ADDR2D(i, j, V)];
		assert(fabs(sum - 1) < MINEPS);
	}
	
	gsl_matrix* M2 = gsl_matrix_alloc(V, V);
	for(int i = 0; i < V; i++)
		for(int j = 0; j < V; j++)
			M2->data[ADDR2D(i, j, V)] = A2->data[ADDR2D(i, j, V)] - (alpha0/(alpha0+1)) * M1->data[i] * M1->data[j];

	printf("time = %lf\n", (double)(clock() - tm_record) / CLOCKS_PER_SEC);

	puts("Whitening M2 ...");
	gsl_matrix* W = whiten(M2, V, K); // W: V x K

	puts("Computing My ...");

	tm_record = clock();
	double Ay0 = 0;
	denom = 1.0 / ndocs;
	for(int i = 0; i < ndocs; i++)
		Ay0 += docs[i].y;
	Ay0 *= denom;
	gsl_vector* Ay1 = gsl_vector_alloc(V);
	gsl_vector_set_all(Ay1, 0);
	for(int i = 0; i < ndocs; i++) {
		int m = docs[i].word_count;
		denom = 1.0 / m;
		for(int j = 0; j < docs[i].item_count; j++) {
			int v = docs[i].index[j];
			Ay1->data[v] += docs[i].count[j] * docs[i].y * denom;
		}
	}
	denom = 1.0 / ndocs;
	for(int i = 0; i < V; i++)
		Ay1->data[i] *= denom;
	gsl_matrix* My = gsl_matrix_alloc(V, V);
	gsl_matrix_set_all(My, 0);
	percentage = 0;
	for(int i = 0; i < ndocs; i++) {
		if (i > ndocs * 0.01 * (percentage + 1)) {
			putchar('.');
			fflush(stdout);
			percentage ++;
		}
		int m = docs[i].word_count;
		double y = docs[i].y;
		denom = 1.0 / (m * (m-1));
		for(int j1 = 0; j1 < docs[i].item_count; j1++)
			for(int j2 = 0; j2 < docs[i].item_count; j2++) {
				int v1 = docs[i].index[j1], v2 = docs[i].index[j2];
				int c1 = docs[i].count[j1], c2 = docs[i].count[j2];
				if (j1 == j2) {
					My->data[ADDR2D(v1, v2, V)] += c1 * (c2-1) * y * denom;
				}
				else {
					My->data[ADDR2D(v1, v2, V)] += c1 * c2 * y * denom;
				}
			}
	}
	puts("");
	denom = 1.0 / ndocs;
	for(int i = 0; i < V; i++)
		for(int j = 0; j < V; j++)
			My->data[ADDR2D(i, j, V)] *= denom;

	for(int i = 0; i < V; i++)
		for(int j = 0; j < V; j++) {
			My->data[ADDR2D(i, j, V)] += -(alpha0/(alpha0+2)) * (Ay0*A2->data[ADDR2D(i, j, V)] + M1->data[i]*Ay1->data[j] + Ay1->data[i]*M1->data[j]) + (2*SQR(alpha0)/((alpha0+1)*(alpha0+2))) * Ay0 * M1->data[i] * M1->data[j];
		}

	gsl_matrix* MyWW = gsl_matrix_alloc(K, K); // MyWW = W' * My * W
	gsl_matrix_set_all(MyWW, 0);
	{
		gsl_matrix* work = gsl_matrix_alloc(V, K);
		gsl_matrix_set_all(work, 0);
		gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, My, W, 0, work);
		gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, W, work, 0, MyWW);
		gsl_matrix_free(work);
	}

	printf("time = %lf\n", (double)(clock() - tm_record) / CLOCKS_PER_SEC);

	puts("Computing 3-rd order moment ...");

	double* M3WWW;
	assert(M3WWW = (double*)malloc(K*K*K * sizeof(double)));
	memset(M3WWW, 0, K*K*K * sizeof(double));

	// E[x1x2x3]
	puts("Part 1 ---");
	tm_record = clock();
	gsl_vector* pw = gsl_vector_alloc(K);
	percentage = 0;
	for(int i = 0; i < ndocs; i++) {
		if (i > ndocs * 0.01 * (percentage + 1)) {
			putchar('.');
			fflush(stdout);
			percentage ++;
		}
		gsl_vector_set_all(pw, 0);

		int m = docs[i].word_count;
		for(int j = 0; j < docs[i].item_count; j++) {
			int v = docs[i].index[j];
			int c = docs[i].count[j];
			for(int k = 0; k < K; k++) {
				pw->data[k] += c * W->data[ADDR2D(v, k, K)];
			}
		}

		double denom1 = 1.0 / m;
		double denom2 = 1.0 / (m-1);
		double denom3 = 1.0 / (m-2);
		for(int k1 = 0; k1 < K; k1++)
			for(int k2 = 0; k2 < K; k2++)
				for(int k3 = 0; k3 < K; k3++)
					M3WWW[ADDR3D(k1, k2, k3, K)] += (pw->data[k1] * denom1) * (pw->data[k2] * denom2) * (pw->data[k3] * denom3); 
	}
	puts("");

	denom = 1.0 / ndocs;
	for(int k1 = 0; k1 < K; k1++)
		for(int k2 = 0; k2 < K; k2++)
			for(int k3 = 0; k3 < K; k3++)
				M3WWW[ADDR3D(k1, k2, k3, K)] *= denom;

	printf("time = %lf\n", (double)(clock() - tm_record) / CLOCKS_PER_SEC);

	puts("Part 2 ---");
	tm_record = clock();
	gsl_matrix* P2 = gsl_matrix_alloc(K, K);
	double *delta;
	assert(delta = (double*)malloc(K*K*K * sizeof(double)));
	percentage = 0;
	for(int i = 0; i < ndocs; i++) {

		if (i > ndocs * 0.01 * (percentage + 1)) {
			putchar('.');
			fflush(stdout);
			percentage ++;
		}
		int m = docs[i].word_count;
		gsl_matrix_set_all(P2, 0);
		gsl_vector_set_all(pw, 0);
		for(int j = 0; j < docs[i].item_count; j++) {
			int v = docs[i].index[j];
			int c = docs[i].count[j];
			for(int k = 0; k < K; k++)
				pw->data[k] += c * W->data[ADDR2D(v, k, K)];
			for(int k1 = 0; k1 < K; k1++)
				for(int k2 = 0; k2 < K; k2++)
					P2->data[ADDR2D(k1, k2, K)] += c * W->data[ADDR2D(v, k1, K)] * W->data[ADDR2D(v, k2, K)];
		}

		denom = 1.0 / (m-2);
		for(int k = 0; k < K; k++)
			pw->data[k] *= denom;
		denom = 1.0 / (m * (m-1));
		for(int k1 = 0; k1 < K; k1++)
			for(int k2 = 0; k2 < K; k2++)
				P2->data[ADDR2D(k1, k2, K)] *= denom; 
		
		for(int k1 = 0; k1 < K; k1++)
			for(int k2 = 0; k2 < K; k2++)
				for(int k3 = 0; k3 < K; k3++) {
					delta[ADDR3D(k1, k2, k3, K)] += pw->data[k1] * P2->data[ADDR2D(k2, k3, K)]
						+ pw->data[k2] * P2->data[ADDR2D(k1, k3, K)]
						+ pw->data[k3] * P2->data[ADDR2D(k1, k2, K)];
				}
			
	}
	puts("");
	denom = 1.0 / ndocs;
	for(int k1 = 0; k1 < K; k1++)
		for(int k2 = 0; k2 < K; k2++)
			for(int k3 = 0; k3 < K; k3++)
				M3WWW[ADDR3D(k1, k2, k3, K)] -= denom * delta[ADDR3D(k1, k2, k3, K)];
	gsl_matrix_free(P2);
	free(delta);

	printf("time = %lf\n", (double)(clock() - tm_record) / CLOCKS_PER_SEC);

	puts("Part 3 ---");
	tm_record = clock();
	gsl_vector* ind = gsl_vector_alloc(V);
	gsl_vector* indw = gsl_vector_alloc(K);
	gsl_vector_set_all(ind, 0);
	int* pind;
	assert(pind = (int*)malloc(ndocs * sizeof(int)));
	memset(pind, 0, ndocs * sizeof(int));
	for(int i = 0; i < V; i++) {
		ind->data[i] = 1;
		gsl_vector_set_all(indw, 0);
		gsl_blas_dgemv(CblasTrans, 1.0, W, ind, 0, indw);
		double coeff = 0;
		for(int j = 0; j < ndocs; j++) {
			int m = docs[j].word_count;
			while (pind[j] < docs[j].item_count && docs[j].index[pind[j]] < i)
				pind[j] ++;
			if (pind[j] < docs[j].item_count && docs[j].index[pind[j]] == i) {
				coeff += (double)(2 * docs[j].count[pind[j]]) / ((double)m * (m-1) * (m-2));
			}
		}
		coeff /= ndocs;
		for(int k1 = 0; k1 < K; k1++)
			for(int k2 = 0; k2 < K; k2++)
				for(int k3 = 0; k3 < K; k3++) {
					M3WWW[ADDR3D(k1, k2, k3, K)] += coeff * indw->data[k1] * indw->data[k2] * indw->data[k3];
				}
		ind->data[i] = 0;
	}
	gsl_vector_free(ind);
	gsl_vector_free(indw);
	free(pind);

	printf("time = %lf\n", (double)(clock() - tm_record) / CLOCKS_PER_SEC);
	// E[M1x1x2], E[x1M1x2], E[x1x2M1], M1^3
	puts("Part 4 ---");
	tm_record = clock();
	gsl_vector* M1W = gsl_vector_alloc(K);
	gsl_vector_set_all(M1W, 0);
	gsl_blas_dgemv(CblasTrans, 1.0, W, M1, 0, M1W); // M1W = M1 * W
	gsl_matrix* A2WW = gsl_matrix_alloc(K, K); //A2WW = W' * A2 * W
	gsl_matrix_set_all(A2WW, 0);
	{
		gsl_matrix* work = gsl_matrix_alloc(V, K);
		gsl_matrix_set_all(work, 0);
		gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, A2, W, 0, work);
		gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, W, work, 0, A2WW);
		gsl_matrix_free(work);
	}
	for(int k1 = 0; k1 < K; k1++)
		for(int k2 = 0; k2 < K; k2++)
			for(int k3 = 0; k3 < K; k3++) {
				M3WWW[ADDR3D(k1, k2, k3, K)] +=	
				  -(alpha0/(alpha0+2)) * (A2WW->data[ADDR2D(k1, k2, K)] * M1W->data[k3] + A2WW->data[ADDR2D(k1, k3, K)] * M1W->data[k2] + A2WW->data[ADDR2D(k2, k3, K)] * M1W->data[k1]) 
				  + (2*SQR(alpha0)/((alpha0+1)*(alpha0+2))) * (M1W->data[k1] * M1W->data[k2] * M1W->data[k3]);
			}
	gsl_matrix_free(A2WW);
	gsl_vector_free(M1W);

	gsl_vector_free(pw);
	gsl_matrix_free(A2);
	gsl_vector_free(Ay1);
	gsl_vector_free(M1);
	gsl_matrix_free(M2);
	gsl_matrix_free(My);

	printf("time = %lf\n", (double)(clock() - tm_record) / CLOCKS_PER_SEC);

	puts("Robust tensor decomposition ...");
	tm_record = clock();
	gsl_vector* lambda = gsl_vector_alloc(K);
	gsl_vector** v;
	assert(v = (gsl_vector**)malloc(K * sizeof(gsl_vector*)));
	for(int k = 0; k < K; k++)
		v[k] = gsl_vector_alloc(K);

	robust_tensor_decomposition(M3WWW, K, V, L, T, lambda, v, rng);

	printf("time = %lf\n", (double)(clock() - tm_record) / CLOCKS_PER_SEC);

	puts("Now recovering sLDA model parameters:");
	strct_slda* slda = alloc_slda(K, V);

	puts("Recovering alpha ...");
	for(int k = 0; k < K; k++)
		slda->alpha->data[k] = 4*alpha0*(alpha0+1)/(SQR(alpha0+2)*SQR(lambda->data[k]));
	// now normalize. NOTE: until this step, alpha must be non-negative
	double sumalpha = gsl_blas_dasum(slda->alpha);
	for(int k = 0; k < K; k++)
		slda->alpha->data[k] = slda->alpha->data[k] / sumalpha * alpha0;

	puts("Recovering mu ...");
	gsl_matrix* Winv = gsl_matrix_alloc(K, V); // Winv = (W'*W)^{-1}* W'; K x V
	gsl_matrix_set_all(Winv, 0);
	{
		gsl_matrix* work = gsl_matrix_alloc(K, K);
		gsl_matrix_set_all(work, 0);
		gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, W, W, 0, work); //work = W' * W
		// compute the inverse of "work"
		gsl_matrix* inv = gsl_matrix_alloc(K, K);
		gsl_permutation* perm = gsl_permutation_alloc(K);
		int s;
		gsl_linalg_LU_decomp(work, perm, &s);
		gsl_linalg_LU_invert(work, perm, inv);
		gsl_blas_dgemm(CblasNoTrans, CblasTrans, 1.0, inv, W, 0, Winv); 
		gsl_matrix_free(work);
		gsl_matrix_free(inv);
		gsl_permutation_free(perm);
	}

	gsl_vector* mu = gsl_vector_alloc(V);
	for(int k = 0; k < K; k++) {
		gsl_vector_set_all(mu, 0);
		gsl_blas_dgemv(CblasTrans, 1.0, Winv, v[k], 0, mu);
		// It is rumored that flipping signs works better than truncating to 0
/*		for(int i = 0; i < V; i++)
			if (mu->data[i] < 0) mu->data[i] = 0;*/
		for(int i = 0; i < V; i++)
			mu->data[i] = fabs(mu->data[i]);
		// normalize
		double summu = gsl_blas_dasum(mu);
		if (summu < MINEPS) {
			for(int i = 0; i < V; i++)
				slda->mu[k]->data[i] = 1.0/V;
		}
		else {
			for(int i = 0; i < V; i++)
				slda->mu[k]->data[i] = mu->data[i] / summu;
		}
	}
	gsl_vector_free(mu);

	// smooth
	for(int k = 0; k < K; k++)
		for(int i = 0; i < V; i++)
			slda->mu[k]->data[i] = (slda->mu[k]->data[i] + 1e-3/V) / (1.0 + 1e-3);

	puts("Recovering eta ...");
	for(int k = 0; k < K; k++) {
		double sum = 0;
		for(int i = 0; i < K; i++)
			for(int j = 0; j < K; j++) {
				sum += ((alpha0+2)/2) * MyWW->data[ADDR2D(i, j, K)] * v[k]->data[i] * v[k]->data[j];
			}
		slda->eta->data[k] = sum;
	}

	gsl_vector_free(lambda);
	for(int k = 0; k < K; k++)
		gsl_vector_free(v[k]);
	free(v);
	gsl_matrix_free(W);
	gsl_matrix_free(Winv);
	gsl_matrix_free(MyWW);
	free(M3WWW);

	return slda;

}
